square(X,Y) :- xmin(X), ymin(Y).
square(X1,Y) :- square(X,Y), X1 = X + 1, xmax(XMax), X1 <= XMax.
square(X,Y1) :- square(X,Y), Y1 = Y + 1, ymax(YMax), Y1 <= YMax.

neighbor(X1,Y,X2,Y) :- square(X1,Y), square(X2,Y), X1 = X2 + 1.
neighbor(X1,Y,X2,Y) :- square(X1,Y), square(X2,Y), X1 = X2 - 1.
neighbor(X,Y1,X,Y2) :- square(X,Y1), square(X,Y2), Y1 = Y2 + 1.
neighbor(X,Y1,X,Y2) :- square(X,Y1), square(X,Y2), Y1 = Y2 - 1.

% The problem-independent encoding in a-star.asp will access the predicates init/1, goal/1, c/4, and h/2 defined here in this problem-specific part of the encoding.
% Predicates action/2 and result/3 are defined in a-star.asp.
init(at(X,Y)) :- start(X,Y).
goal(at(X,Y)) :- goal(X,Y).
c(at(AtX,AtY),move(X,Y),at(X,Y),StepCost) :- neighbor(AtX,AtY,X,Y), not obstacle(X,Y), StepCost=1.
h(at(X,Y),H) :- manhattan(X,Y,H).

% five separate rules to compute Manhattan distances of individual tiles because Alpha currently does not support computing absolute values:
xdist(X,Y,XDist) :- square(X,Y), goal(GoalX,GoalY), XDist=GoalX-X, XDist >= 0.
xdist(X,Y,XDist) :- square(X,Y), goal(GoalX,GoalY), XDist=X-GoalX, XDist > 0.
ydist(X,Y,YDist) :- square(X,Y), goal(GoalX,GoalY), YDist=GoalY-Y, YDist >= 0.
ydist(X,Y,YDist) :- square(X,Y), goal(GoalX,GoalY), YDist=Y-GoalY, YDist > 0.
manhattan(X,Y,Manhattan) :- xdist(X,Y,XDist), ydist(X,Y,YDist), Manhattan=XDist+YDist.
