elem(z,Z) :- zone2sensor(Z,D).
elem(s,D) :- zone2sensor(Z,D).

elems_assigned_up_to(0).
assignable_unit(1).
assignable_unit(U) :- comUnit(U), used(U-1).
assignable(U,T,X) :- assignable_unit(U), elem(T,X), elem_n(T,X,N), elems_assigned_up_to(N-1).

{ assign(U,T,X) } :- assignable(U,T,X).
#heuristic F assign(U,T,X) : assignable(U,T,X), full(U,T), maxLayer(ML), numUnits(NU). [ (ML+1)*NU ] % do not use full units
#heuristic   assign(U,T,X) : assignable(U,T,X), maxLayer(ML), layer(T,X,L), used(U), Level=ML-L, numUnits(NU). [ Level*NU + U ] % try preceding units in decreasing order
#heuristic   assign(U,T,X) : assignable(U,T,X), maxLayer(ML), layer(T,X,L), comUnit(Um1), used(Um1), not assign(Um1,T,X), Um1=U-1, Level=ML-L, numUnits(NU). [ Level*NU + U ] % use the next fresh unit after all preceding units have been tried

% disable all other units when one is used:
:- comUnit(U), comUnit(OtherU), elem(T,X), assign(U,T,X), assign(OtherU,T,X), U != OtherU.

% "do not use full units" as a constraint:
:- full(U,T), assign(U,T,X1), assign(U,T,X2), X1<X2, elem(T,X3), X3 != X1, X3 != X2, assign(U,T,X3).

:- assign(U,T,X1), assign(U,T,X2), assign(U,T,X3), X1<X2, X2<X3.
%:- assign(U1,T,X), assign(U2,T,X), U1<U2. % covered by "do not use full units" constraint above
:- elem(T,X), not assigned(T,X).

partnerunits(U,P) :- assign(U,z,Z), assign(P,s,D), zone2sensor(Z,D), U!=P.
partnerunits(U,P) :- partnerunits(P,U).
:- partnerunits(U,P1), partnerunits(U,P2), partnerunits(U,P3), P1<P2, P2<P3.

% helpers for QUICKPUP heuristics, inspired by quickpup.ascass by Erich Teppan:
sensor(S) :- zone2sensor(Z,S).
zone(Z) :- zone2sensor(Z,S).

numZones(N) :- zone(N), not zone(N+1).
numSensors(N) :- sensor(N), not sensor(N+1).
numElems(E+F) :-numZones(E), numSensors(F).
numUnits(N) :- comUnit(N), not comUnit(N+1).

%-------topological order---
startZone(1).	% future work: other startZones (random choice?)
zoneDist(Z0,0) :- startZone(Z0).

sensorDist(S,Dz+1) :- zoneDist(Z,Dz), zone2sensor(Z,S), numElems(M), Dz<M.
zoneDist(Z,Ds+1) :- sensorDist(S,Ds), zone2sensor(Z,S), numElems(M), Ds<M.

zoneLayer(Z,Dmin) :- zone(Z), zoneDist(Z,Dmin), not zoneDist(Z,Dmin-2).
sensorLayer(S,Dmin) :- sensor(S), sensorDist(S,Dmin), not sensorDist(S,Dmin-2).

layer(L) :- zoneLayer(_,L).
layer(L) :- sensorLayer(_,L).
maxLayer(N) :- layer(N), not layer(N+1).

full(U,T) :- comUnit(U), assign(U,T,X1), assign(U,T,X2), X1<X2.
assigned(T,X) :- assign(_,T,X).
used(U) :- comUnit(U), assign(U,_,_).

% total ordering over all elements (can be fully evaluated by stratified evaluation):
elementLayer(z,Z,L) :- zoneLayer(Z,L).
elementLayer(s,S,L) :- sensorLayer(S,L).
elem_n(z,Z0,1) :- startZone(Z0).
elem_n(T2,X2,N+1) :- elem_n(T1,X1,N), elementLayer(T1,X1,L1), max_element_on_layer(T1,X1), elementLayer(T2,X2,L1+1), min_element_on_layer(T2,X2).
elem_n(T,X2,N+1) :- elem_n(T,X1,N), elementLayer(T,X1,L), elementLayer(T,X2,L), X1 < X2, not exists_element_between_on_layer(T,X1,X2,L).
min_element_on_layer(T,X) :- elementLayer(T,X,L), not exists_lower_element_on_layer(T,X,L).
exists_lower_element_on_layer(T,X,L) :- elementLayer(T,X,L), elementLayer(T,LowerX,L), LowerX < X.
max_element_on_layer(T,X) :- elementLayer(T,X,L), not exists_higher_element_on_layer(T,X,L).
exists_higher_element_on_layer(T,X,L) :- elementLayer(T,X,L), elementLayer(T,HigherX,L), HigherX > X.
exists_element_between_on_layer(T,X1,X2,L) :- elementLayer(T,X1,L), elementLayer(T,X,L), elementLayer(T,X2,L), X1 < X, X < X2.

% which elements (in the total order) are already assigned:
elems_assigned_up_to(1) :- elem_n(T,X,1), assigned(T,X).
elems_assigned_up_to(N) :- elems_assigned_up_to(N-1), elem_n(T,X,N), assigned(T,X).

% translate old breadth-first syntax to new one:
layer(z,X,L) :- zoneLayer(X,L).
layer(s,X,L) :- sensorLayer(X,L).

% output format:
unit2zone(U,Z)   :- assign(U,z,Z).
unit2sensor(U,S) :- assign(U,s,S).
% partnerunits/2 is already defined above
